import sys
import re

from gawseed.threatfeed.search import Search

class RESearch(Search):
    """A base class for regular expression based search engines to reuse functions from.

       `search_items` should be a list regular expressions, and will be built
        into self._relist as {'match': item, 're': compiled_version }"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)
        self.require(['key'])
        self._key = self.config('key', help="The key field to match against in the data stream")

        self._search_list = search_list
        self._data_iterator = data_iterator
        self._binary_search = binary_search

    def initialize(self):
        super().initialize()
        self._relist = []
        for item in self._search_list:
            try: 
                compiled = re.compile(self._data_iterator.decode_item(item))
                self._relist.append({ 'match': self._search_list[item],
                                      're': compiled})
            except:
                sys.stderr.write("failed to compile regular expression: %s\n" % (item))
    
    def search(self, row):
        """If the `source` value matches any stored expression, 
           the expression will be returned."""
        return self.search_one(self._data_iterator.decode_item(row[self._key]))

    def search_one(self, source):
        """If the `source` value matches any stored expression, 
           the expression will be returned."""
        if source == None:
            return
        for expr in self._relist:
            match = expr['re'].match(source)
            if match:
                return expr['match']
            



from gawseed.threatfeed.config import Config

class Search(Config):
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf)
        self._search_list = search_list
        self._data_iterator = data_iterator
        self._binary_search = binary_search

    def initialize(self):
        self._search_list = self._data_iterator.encode_dict(self._search_list)
    
    def __iter__(self):
        return self

    def __next__(self):
        for row in self._data_iterator:
            match = self.search(row)
            if match:
                row = self._data_iterator.convert_row_to_utf8(row)
                yield (row, match)


        



from collections import Counter
import dnssplitter

from gawseed.threatfeed.config import Config

from .summarizer import Summarizer

class DNSSummarizer(Summarizer):
    """Summarizes/counts data from dns requests from another enrichment
    module, counting the results by full query (into 'queries') and by
    domains and regpoints (in 'domains' and 'regpoints') respectively.
    """
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf, search_index, dataset, is_binary, loader)

        self._splitter = dnssplitter.DNSSplitter()
        self._splitter.init_tree()
        
    def gather(self, count, row, match, enrichment_data):
        """Re-sort all the enrichment data based on the specified column"""
        if not self.check_enrichment_data(enrichment_data):
            return (None, None)

        query_counter = Counter()
        domain_counter = Counter()
        regpoint_counter = Counter()
        for row in enrichment_data[self._enrichment_key]:
            if self._data_key in row:
                query_counter[row[self._data_key]] += 1
                split_dns = self._splitter.search_tree(row[self._data_key])
                if split_dns:
                    (prefix, domain, regpoint) = split_dns
                    domain_counter[domain] += 1
                    regpoint_counter[regpoint] += 1

        return (self._output_key, {
            'queries': dict(query_counter),
            'domains': dict(domain_counter),
            'regpoints': dict(regpoint_counter),
        })
                



#!/usr/bin/python3

"""This is a generic threat-feed searching application that takes a
threat feed of various types and matches it against sources of network
log data to search through.  It can take both threat-feeds and log
data from multiple sources, including kafka and files.  It's output is
configurable and can be given a jinja2 template for generic report
generation.
"""

import sys, os
import re
import time
import argparse
from copy import deepcopy

import datetime
import importlib
import traceback

import multiprocessing
import threading
import queue

from msgpack import unpackb

from gawseed.threatfeed.loader import Loader

debug = False

loader = None

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description="This uses data from threat feeds (in kafka topics or files) to search through network data (from kafka or files) and reports any matches as 'events'.")

    # THREAT FEEDS
    # ----------------------------------------------------------------------
    # arguments for the threat feed source
    group = parser.add_argument_group("Threat feed arguments")
    group.add_argument("-B", "--threat-begin-time", default=None, type=str,
                        help="Date to start returning threat feed data from")

    group.add_argument("-E", "--threat-end-time", default=None, type=str,
                        help="Date to stop returning threat feed data after")

    group.add_argument("-M", "--threat-max-records", default=None, type=int,
                        help="Maximum number of threat records to retrieve and search for ")

    group.add_argument("--threat-timeout", default=1000, type=int,
                        help="Maximum number of ms to wait for a threat feed before figuring its done")


    # arguments for kafka threat feeds
    group = parser.add_argument_group("Kafka specific threat feed arguments")
    group.add_argument("-K", "--threat-kafka-servers", default=[], type=str,
                        nargs="*", help="Kafka server to pull topics from")

    group.add_argument("--threat-kafka-topic", default="gawseed_ipaddresses", type=str,
                        help="The kafka threat topic to use as the threat source")

    group = parser.add_argument_group("File/FSDB specific threat feed arguments")
    # arguments for file based threat feeds
    group.add_argument("--threat-fsdb", default=None, type=str,
                        help="FSDB based threat feed")

    # DATA FEEDS
    # ----------------------------------------------------------------------
    # arguments for the data feeds
    group = parser.add_argument_group("Data feed arguments")
    group.add_argument("-T", "--time-column", default="ts", type=str,
                        help="The time column to read looking for end-time")

    group.add_argument("-m", "--max-records", type=int,
                        help="maximum number of records to return")

    group.add_argument("-b", "--begin-time", default=None, type=str,
                        help="Date to start returning threat feed data from")

    group.add_argument("-e", "--end-time", default=None, type=str,
                        help="Date to stop returning threat feed data after")

    # kafka specific data feed options
    group = parser.add_argument_group("Kafka specific data feed arguments")
    group.add_argument("--data-kafka-servers", default=[], type=str,
                        nargs="*", help="Kafka server to pull data streams from")

    group.add_argument("-t", "--data-topic", default="ssh", type=str,
                        help="Kafka topic to request from the data source")

    group = parser.add_argument_group("BRO specific data feed arguments")
    # BRO file
    group.add_argument("--bro-data", type=str,
                        help="Bro file to read data from")

    group = parser.add_argument_group("File/FSDB specific data feed arguments")
    # fsdb data specific options
    group.add_argument("--fsdb-data", default=None, type=str,
                        help="FSDB data file for testing")


    group = parser.add_argument_group("Output formatting arguments")
    # OUTPUT
    # ----------------------------------------------------------------------
    # output processing
    group.add_argument("-f", "--fields", default=None, type=str, nargs="*",
                        help="Fields to print")
    
    group.add_argument("--fsdb", action="store_true",
                        help="Output FSDB formatted data")

    group.add_argument("-o", "--output-pattern", default=None, type=str,
                        help="Output a pattern of files instead of stdout; include %%d for a count")

    group.add_argument("-j", "--jinja-template", default=None, type=str,
                        help="The jinja template to use when generating reports")

    group.add_argument("-J", "--jinja-extra-information", default=None, type=str,
                        help="Extra information in YAML format to include with report generation in 'extra' an field")
    parser.add_argument("--threads", default=2, type=int,
                        help="Number of output stream threads to create.")

    # Configuration
    group = parser.add_argument_group("Global configuration")
    group.add_argument("-y", "--config", type=argparse.FileType("r"),
                       help="A YAML configuration file specifying all modules to be loaded")
    group.add_argument("-Y", "--config-parameters", type=str, nargs="*", default=[],
                       help="A list of parameters to pass to the YAML config as jinja template variables.  The parameters should be in the form of name=value pairings")

    # DEBUGGING
    group = parser.add_argument_group("Debugging arguments")
    group.add_argument("--dump-threat-feed", action="store_true",
                        help="Simply dump the threat feed data to stdout")
   
    group.add_argument("--dump-data", action="store_true",
                        help="Simply dump all the data without searching to stdout")

    group.add_argument("--dump-events", action="store_true",
                        help="Simply dump all the event data in raw form")

    group.add_argument("--config-templates", action="store_true",
                       help="Dump all module configuration options")

    group.add_argument("-V", "--verbose", action="store_true",
                        help="Verbose/Debugging output")

    args = parser.parse_args()
    if args.verbose:
        global debug
        debug = True

    if args.config_templates:
        loader.dump_config_options(debug)

    # if args.merge_grep and not args.merge_key:
    #     raise ValueError("--merge-key/-k is required with --merge-grep")

    return args

def verbose(number, msg=None):
    if not msg:
        msg = number
        number = ""
    else:
        msg = "ps " + str(number) + ": " + msg

    if debug:
        print(msg)

def get_threat_feed(number, args, conf=None, max_records=None, dump=False):
    """Read in the threat feed stream as a data source to search for"""

    threat_source = loader.create_instance_for_module(conf, loader.THREATSOURCE_KEY)

    verbose(number, "created threat feed: " + str(threat_source))

    # initialize and read
    threat_source.open()
    (search_data, search_index) = threat_source.read(max_records)

    verbose(number, "  read feed with " + str(len(search_data)) + " entries")

    if dump:
        import json
        print(json.dumps(search_data))
        print(json.dumps(search_index))
        exit(0)

    return (threat_source, search_data, search_index)


def get_data_source(number, args, conf=None):
    """Get the data source and open it for traversing"""

    data_source = loader.create_instance_for_module(conf, loader.DATASOURCE_KEY)

    verbose(number, "created data feed: " + str(data_source))

    # just print it?
    if args.dump_data:
        import json
        for finding in data_source:
            print(json.dumps(data_source.convert_row_to_utf8(finding)))
        exit(0)

    return data_source

def get_searcher(number, args, search_index, data_source, conf=None):
    """Create a searcher object"""
    # create the searching interface
    searcher = loader.create_instance_for_module(conf, loader.SEARCHER_KEY,
                                                 [search_index, data_source,
                                                  data_source.is_binary()])
    verbose(number, "created searcher: " + str(searcher))

    return searcher

def get_outputs(number, conf):
    """Create the output-er object"""

    outputs = []
    section = conf[loader.REPORTER_KEY]

    if type(section) is not list:
        section = [section]

    for item in section:
        obj = loader.create_instance(item, loader.REPORTER_KEY)
        outputs.append(obj)

    return outputs

def get_enrichments(number, conf, search_index, data_source):
    if loader.ENRICHMENT_KEY not in conf:
        return []
    section = conf[loader.ENRICHMENT_KEY]
    enrichers = []

    for item in section:
        enricher = loader.create_instance(item, loader.ENRICHMENT_KEY,
                                          [search_index, data_source,
                                           data_source.is_binary(), loader])
        enrichers.append(enricher)
    return enrichers

def found_event(found_queue, enrichers, outputs):
    while True:
        event = found_queue.get()
        if event is None:
            break # None in queue signals an exit

        row = event['row']
        match = event['match']
        count = event['count']

        enrichment_data = {}

        # gather enrichment data from the backends
        for ecount, enricher in enumerate(enrichers):
            try:
                (key, result) = enricher.gather(count, row, match, enrichment_data)
                if key and result:
                    enrichment_data[key] = result
            except Exception as e:
                sys.stderr.write("An enricher failed: " + str(e) + "\n")
                sys.stderr.write("".join(traceback.format_exc()))
                sys.stderr.write(str(enricher.get_config()) + "\n")
                if 'errors' not in enrichment_data:
                    enrichment_data['errors'] = []
                enrichment_data['errors'].append({ 'count': count,
                                                   'module': type(enricher),
                                                   'msg': 'An enrichment module failed to load data'})

        # loop through the outputs to create stuff
        for output in outputs:
            try:
                output.new_output(count)
                output.write(count, row, match, enrichment_data)
                output.maybe_close_output()
                
            except Exception as e:
                sys.stderr.write("The output module failed: " + str(e) + "\n")
                sys.stderr.write("".join(traceback.format_exc()))
        

        # signal this entry is done
        found_queue.task_done()

def convert_args_to_config(args):
    subconf = {}

    # ----- threat feed
    # read in the threat feed stream as a data source to search for
    conf_part = {}
    if args.threat_fsdb:
        conf_part = { 'module': 'fsdb',
                      'file': args.threat_fsdb}
    elif args.threat_kafka_servers:
        conf_part = { 'module': 'kafka',
                      'bootstrap_servers': args.threat_kafka_servers,
                      'begin_time': args.threat_begin_time,
                      'topic': args.threat_kafka_topic,
                      'partition': 0,
                      'timeout': args.threat_timeout}
    else:
        raise ValueError("no data source specified")

    subconf[loader.THREATSOURCE_KEY] = conf_part

    # ----- data source
    conf_part = {}
    if args.fsdb_data:
        conf_part = { 'module': 'fsdb',
                      'file': args.fsdb_data }
    elif args.bro_data:
        conf_part = { 'module': 'bro',
                      'file': args.bro_data }
    elif args.data_kafka_servers:
        conf_part = { 'module': 'kafka',
                      'bootstrapservers': args.data_kafka_servers,
                      'begin_time': args.begin_time,
                      'topic': args.data_topic }
    else:
        raise ValueError("no data source specified")
    subconf[loader.DATASOURCE_KEY] = conf_part

    # ----- searcher
    conf_part = {}
    if args.data_topic in ['ssh', 'http']:
        conf_part = {'module': args.data_topic}
    elif args.data_topic in ['ip', 'conn']:
        conf_part = {'module': 'ip'}
    else:
        raise ValueError("no searcher specified")

    subconf[loader.SEARCHER_KEY] = conf_part

    # ----- outputs
    conf_part = {}
    if args.dump_events:
        conf_part = {'module': 'dumper',
                     'stream': args.output_pattern}
    elif args.jinja_template:
        conf_part = {'module': 'reporter',
                     'stream': args.output_pattern,
                     'template': args.jinja_template,
                     'extra_information': args.jinja_extra_information
        }
    else:
        conf_part = {'module': 'reporter' } # default
    subconf = [loader.REPORTER_KEY] = conf_part


    conf = {loader.YAML_KEY: [subconf]}
    return conf

def launch_process(combination, args, number):
    """Creates a process to handle a given threat-search configuration entry"""
    # pass in verbosity level
    if args.verbose:
        for subsection in combination:
            item = combination[subsection]
            if type(item) == list:
                for i in item:
                    i['verbose'] = True
            else:
                item['verbose'] = True
                

    (threat_source, search_data, search_index) = \
        get_threat_feed(number, args, combination,
                        args.threat_max_records, args.dump_threat_feed)

    data_source = get_data_source(number, args, combination)
    searcher = get_searcher(number, args, search_index, data_source, combination)

    data_source.open()

    enrichers = get_enrichments(number, combination, search_index, data_source)

    outputs = get_outputs(number, combination)
    verbose(number, "created outputs: " + str(outputs))

    # loop through all the data for matches
    if debug:
        print("reports created: 0", end="\r")

    # create threads to handle the results
    event_queue = queue.Queue()
    output_threads = []
    for i in range(args.threads):
        thread = threading.Thread(target=found_event,
                                  args=(event_queue, enrichers, outputs))
        thread.start()
        output_threads.append(thread)

    # 
    for count, results in enumerate(next(searcher)):
        row, match = results

        event_queue.put(
            { 'row': deepcopy(row),
              'match': deepcopy(match),
              'count': count,
            })

        if debug:
            print("%d: events found: %d" % (number, count+1), end="\r")

        if args.max_records and count >= args.max_records:
            break

    # tell the threads to quit
    for n in range(args.threads):
        event_queue.put(None)

    # wait for threads to clear
    for t in output_threads:
        t.join()


def main():
    # create our loading class
    global loader
    loader = Loader()

    args = parse_args()

    if args.config:
        conf = loader.load_yaml_config(args.config, args.config_parameters)
    else:
        conf = convert_args_to_config

    threat_conf = conf[loader.YAML_KEY]

    processes = []

    if len(threat_conf) == 1:
        # don't bother with all that subprocessing
        launch_process(threat_conf[0], args, 0)
        verbose("")
        return

    for number, combination in enumerate(threat_conf):
        subprocess = multiprocessing.Process(target=launch_process,
                                             args=(combination, args, number))
        subprocess.start()
        processes.append(subprocess)

    for process in processes:
        process.join()
    verbose("")

if __name__ == "__main__":
    main()



from io import StringIO
from gawseed.threatfeed.events.reporter import EventStreamReporter

class RedmineReporter(EventStreamReporter):
    """Takes a markdown output and sends the results to a redmine
    instance."""

    def __init__(self, conf):
        super().__init__(conf)

        self.require(["redmine_host"])
        self._redmine_host = self.config("redmine_host",
                                         help="Redmine host to connect to")

    def new_output(self, count, **kwargs):
        self._stream = StringIO()
        

    def maybe_close_output(self):
        pass # send to redmine here
    



import json
import time
import sys
import traceback

from gawseed.threatfeed.config import Config

class Datasource(Config):
    """Pulls enrichment data from another datasource based on a key
    ("match_key") in the row that was matched by the searcher.  It
    matches this key against a key from the datasource:
    "datasource_key.  When a match is found, data will be gathered
    backward and forward from that time based on the specified
    backward_time and forward_time values.  The set of rows matched
    will be returned as an array.
    """
    def __init__(self, conf, search_index, data, is_binary, loader=None):
        super().__init__(conf)

        self.require(['datasource'])

        self._datasource = self.config('datasource',
                                       help="A dictionary describing a datasource to pull data from.")
        self._match_key = self.config('match_key', 'id_orig_h',
                                      help="The key to use when extracting search data from the matched row.")

        self._datasource_time_column = self.config('datasource_time_column', 'timestamp',
                                          help="The time column to use when extracting a match timestamp to search forward and backward from")
        
        self._datasource_key = self.config('datasource_key', 'id_orig_h',
                                           help="The key to use when searching through the datasource.")

        self._time_backward = self.config('time_backward', '2m', datatype='offset',
                                          help="The amount of time to get data from going backward from the match point")
        self._time_forward = self.config('time_forward', '2m', datatype='offset',
                                         help="The amount of time to get data from going forward from the match point")

        self._output_key = self.config('output_key', 'datasource',
                                       help="The output key to store the returned data in.")

        self._loader = loader

    def gather(self, count, row, match, enrichment_data = {}):

        ds_config = self._datasource

        timeval = row[self._datasource_time_column]
        timestamp = self.parse_time(timeval)
        ds_config['begin_time'] = '@' + str(timestamp - self._time_backward)
        ds_config['end_time'] = '@' + str(timestamp + self._time_forward)

        # have the loader create the data_source from the config
        data_source = self._loader.create_instance(ds_config,
                                                   self._loader.DATASOURCE_KEY)
        data_source.initialize()

        # Have the loader create a searcher to search the data source
        conf = { 'module': 'ip',
                 'search_keys': [self._datasource_key]}

        search_index = {row[self._match_key]: row}
        searcher = self._loader.create_instance(conf, self._loader.SEARCHER_KEY,
                                                [search_index, data_source,
                                                 data_source.is_binary()])

        # tell the datasource to start up
        try:
            data_source.open()
        except Exception as e:
            print("no data at start -- end of file? " + str(e))
            return (self._output_key, []) # end of file

        self.verbose("enrichment/datasource searcher created")
        self.verbose("  searching from " + str(ds_config['begin_time']) + " to " + str(ds_config['end_time']))
        self.verbose("  " + str(self.get_config()))

        # collect everything from the datasource into a row 
        enrich_rows = []
        try:
            for finding in next(searcher):
                enrich_rows.append(finding[0])
        except Exception as e:
            print(traceback.format_exc())
            print("done searching? exception: " + str(e))

        self.verbose("  found " + str(len(enrich_rows)) + " rows for key=" + self._output_key)

        # return the key, and the found rows
        return (self._output_key, enrich_rows)
                



import sys
import json

from gawseed.threatfeed.feeds import ThreatFeed

from kafka import KafkaConsumer
from kafka.structs import TopicPartition

class KafkaThreatFeed(ThreatFeed):
    """Pulls threat data from a GAWSEED project (or other) kafka threat-feed source."""
    def __init__(self, conf):
        super().__init__(conf)
        self.require(['bootstrap_servers', 'topic', 'partition'])

        self._bootstrap_servers = self.config('bootstrap_servers',
                                              help="A list of kafka bootstrap servers to query")
        self._topic = self.config('topic',
                                  help="The kafka topic to stream from")
        self._partition = self.config('partition',
                                      help="The kafka partition to stream from")
        self._timeout = self.config('timeout',
                                                help="A timeout in milliseconds to wait for server data.")
        self._max_records = self.config('limit',
                                        help="Maximum number of records to retrieve")

    def open(self):
        super().initialize()
        
        self._consumer = KafkaConsumer(bootstrap_servers=self._bootstrap_servers,
                                       consumer_timeout_ms=self._timeout)

        # point to what we want at
        partition = TopicPartition(self._topic, self._partition)
        self._consumer.assign([partition])

        offset = None
        if self._begin_time:
            timestamp = self._begin_time * 1000
            offinfo = self._consumer.offsets_for_times({partition: timestamp})
            if offinfo == None or offinfo[partition] == None:
                raise ValueError("There is no data in the threat feed stream the begin date")
            offset = offinfo[partition].offset
            self._consumer.seek(partition, offset)

    def parse_record(self, record):
        entry = json.loads(record.value)
        return entry

    def next_row(self):
        return next(self._consumer)

    def read(self, max_records=None, remove_duplicates=True):
        array = []
        dictionary = {}
        if not max_records:
            max_records = self._max_records

        if self._begin_time:
            timestamp = self._begin_time

        for (count, entry) in enumerate(self._consumer):
            try:
                entry = self.parse_record(entry)

                # tmp hack to work around kafka hanging on some topics;
                # thus we start from the beginning and read everything.
                # this naturally won't scale.
                if self.maybe_drop_entry(entry, self._value_column):
                    continue
                
                # don't duplicate signatures if requested not to
                if remove_duplicates and entry[self._value_column] in dictionary:
                    continue
                
                dictionary[entry[self._value_column]] = entry # note, may erase older ones; build array?
                array.append(entry)
            except Exception as e:
                sys.stderr.write("dropping kafka feed entry due to a parse error: " + str(entry) + "\n")
                sys.stderr.write(str(e) + "\n")
                
            if max_records and count >= max_records:
                break

        return (array, dictionary)




from concurrent.futures import ProcessPoolExecutor, wait, FIRST_COMPLETED

from gawseed.threatfeed.search import Search
from gawseed.threatfeed.loader import Loader

class Parallel(Search):
    """Searches for IP address threats in multiple search_keys fields"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)
        self._processes = self.config("processes", default=4,
                                       help="Number of processes to spawn in parallel")
        self._preload = self.config("preload",
                                    help="Number of records to preload into the queue for the processors (the default will be four times the number of processes if left unset)")

        self._searcher_conf = self.config("searcher",
                                     help="A dictionary key describing the searcher to use")

        self._loader = Loader()
        self._search_list = search_list
        self._data_source = data_iterator
        self._binary_search = binary_search

    def initialize(self):
        super().initialize()

        if not self._preload:
            self._preload = 4 * self._processes

        # create queue
        self._futures = {}
        self._futures_num = 0
        self._answers = []

        # create the executor pool
        self._pool = ProcessPoolExecutor(max_workers=self._processes,
                                         initializer=search_init,
                                         initargs=(self._loader,
                                                   self._searcher_conf,
                                                   self._search_list,
                                                   self._data_source,
                                                   self._binary_search))

    def shutdown(self):
        self._pool.shutdown()

    def __next__(self):
        running = True
        while running:
            # load up the futures queue with PRELOAD items from the data source
            while len(self._futures) < self._preload:
                # get the next row
                try:
                    row = next(self._data_iterator)
                except Exception as e:
                    # on stop iteration of the datasource,
                    # we should stop running
                    # XXX: really need to keep looping to catch ending events 
                    running = False
                    
                if not row:
                    continue

                # our data is not full, so just load an item
                self._futures_num += 1
                self._futures[self._futures_num] = \
                    self._pool.submit(search_one, (self._futures_num, row))

            # queue is full, so wait for a searcher to be done
            if len(self._futures) != 0:
                wait(self._futures.values(), return_when=FIRST_COMPLETED)
            else:
                self.verbose("stuck???")

            # process anything returned, deleting handled instances
            delete_these = []
            for f in self._futures:
                if self._futures[f].done():
                    (num, row, answer) = self._futures[f].result()
                    if answer:
                        self._answers.append((row, answer))
                    delete_these.append(num)

            for d in delete_these:
                del self._futures[d]

            if len(self._answers) > 0:
                (row, match) = self._answers.pop(0)
                yield (row, match)

        self.shutdown()
    

# sub-process searching / state

searcher=None
var=None

def search_init(loader, conf, search_index, data_source, binary_search):
    global searcher
    import random
    var = random.randint(0,1000)
    data_source.close()
    searcher = \
        loader.create_instance(conf, loader.SEARCHER_KEY,
                               [search_index,
                                data_source,
                                binary_search])
    
def search_one(args):
    (num, row) = args
    match = searcher.search(row)
    return (num, row, match)




#!/usr/bin/python3

"""Reads a gawseed-threat-feed pickle file and pipes it to a reporter
for generating reports from archived data"""

import pickle
from gawseed.threatfeed.loader import Loader

import argparse
import sys

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description="""Reads a gawseed-threat-feed pickle file and pipes it to a reporter for generating reports from archived data""")

    parser.add_argument("-j", "--jinja-template", default=None, type=str,
                        help="The jinja template to use when generating reports")

    parser.add_argument("-J", "--jinja-extra-information", default=None, type=str,
                        help="Extra information in YAML format to include with report generation in 'extra' an field")

    parser.add_argument("--json", default=None, type=argparse.FileType('rb'),
                        help="An extra json file to read")

    parser.add_argument("--json-extra-name", default="test", type=str,
                        help="The tag name to load the json data into")

    parser.add_argument("pickle_file", type=argparse.FileType('rb'),
                        nargs='?', default=sys.stdin,
                        help="The input pickle archive file to load")

    args = parser.parse_args()

    if not args.jinja_template:
        raise ValueError("-j is a required argument")

    return args

def main():
    args = parse_args()

    data = pickle.loads(args.pickle_file.read())
    conf = { 'module': 'reporter',
             'template': args.jinja_template,
             'extra_information': args.jinja_extra_information}

    if args.json:
        import json
        data[args.json_extra_name] = json.loads(args.json.read())
        data['enrichments'][args.json_extra_name] = data[args.json_extra_name]

    loader = Loader()
    reporter = loader.create_instance(conf, loader.REPORTER_KEY)
    reporter.new_output(0)
    reporter.write(0, data['row'], data['match'], data['enrichments'])
    reporter.maybe_close_output()

if __name__ == "__main__":
    main()




from gawseed.threatfeed.config import Config

class ThreatFeed(Config):
    """Base class for threat feeds"""
    def __init__(self, conf):
        super().__init__(conf)

        self._begin_time = self.config('begin_time', datatype='time',
                                       help="The time to start searching from; no value will mean end of stream")
        self._end_time = self.config('end_time', datatype='time',
                                     help="The time to stop a search when reading; no value will mean don't stop streaming")
        self._time_column = self.config('time_column', default='timestamp',
                                        help="Time column to use when searching through data")

        self._value_column = self.config('key', 'value',
                                         help="The primary column/key name to use for pulling threat data")

        self._exclude_list = self.config('exclude', [],
                                         help='A list of entries to ignore in the threat feed')

    def initialize(self):
        if type(self._exclude_list) != list:
            self._exclude_list = [self._exclude_list]

    def __iter__(self):
        return self

    def __next__(self):
        row = self.next_row(self._consumer)
        if self._end_time and row[self._time_column] >= self._end_time:
            raise StopIteration()
        return row

    def maybe_drop_entry(self, entry, value_column):
        if self._begin_time and int(float(entry[self._time_column])) < self._begin_time:
            return True
        
        if value_column not in entry:
            return True

        if entry[value_column] in self._exclude_list:
            return True

        return False
        



from gawseed.threatfeed.config import Config

class ConnectionCounter(Config):
    """Summarizes connection information in BRO or similar data"""
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf)

        self.require(['enrichment_key', 'output_key'])

        self._orig_key = self.config('origin_key', 'id_orig_h',
                                     help="The origin key")
        self._resp_key = self.config('destination_key', 'id_resp_h',
                                     help="The destination key")
        self._port_key = self.config('destination_port', 'id_resp_p',
                                     help="The destination port to group by")
        self._orig_bytes_key = self.config('origin_bytes_key', 'orig_bytes',
                                           help="The key for the bytes sent by the origin ")
        self._resp_bytes_key = self.config('destination_bytes_key',
                                           'resp_bytes',
                                           help="The key for the bytes sent by the server ")
        self._enrichment_key = self.config("enrichment_key",
                                           help="The enrichment key for the data to be analyzed")
        self._output_key = self.config('output_key', 'datasource',
                                       help="The output key to store the analysis data in")

    def gather(self, count, row, match, enrichment_data):
        """Collect and tally results of all the connections"""
        # extract the current data
        if self._enrichment_key not in enrichment_data:
            return (None, None)

        conns = {}
        ports = {}
        for row in enrichment_data[self._enrichment_key]:
            orig = row[self._orig_key]
            resp = row[self._resp_key]
            port = row[self._port_key]
            rxbytes = row[self._orig_bytes_key]
            txbytes = row[self._resp_bytes_key]

            if rxbytes is None or rxbytes == "-":
                rxbytes = 0
            else:
                rxbytes = int(rxbytes)
            if txbytes is None or txbytes == "-":
                txbytes = 0
            else:
                txbytes = int(txbytes)

            base_info = { 'count': 1,
                          'rxbytes': rxbytes,
                          'txbytes': txbytes }
            if orig not in conns:
                conns[orig] = {resp: {port: base_info}}
                continue

            if resp not in conns[orig]:
                conns[orig][resp] = {port: base_info}
                continue

            if port not in conns[orig]:
                conns[orig][resp][port] = base_info
            else:
                conns[orig][resp][port]['count'] += 1
                conns[orig][resp][port]['rxbytes'] += rxbytes
                conns[orig][resp][port]['txbytes'] += txbytes

            if port not in ports:
                ports[port] = 1
            else:
                ports[port] += 1
                
        results = {'connections': conns,
                   'ports': ports}
        return (self._output_key, results)



import importlib
import re
import yaml
import traceback
import jinja2

YAML_KEY = 'threat-search'
THREATSOURCE_KEY = 'threatsource'
DATASOURCE_KEY = 'datasource'
SEARCHER_KEY = 'searcher'
REPORTER_KEY = 'reporter'
ENRICHMENT_KEY = 'enrichments'
YAML_SECTIONS = [THREATSOURCE_KEY, DATASOURCE_KEY, SEARCHER_KEY, REPORTER_KEY, ENRICHMENT_KEY]

# load modules?
MODULE_XFORMS = {
    THREATSOURCE_KEY: {
        'kafka': 'gawseed.threatfeed.feeds.kafka.KafkaThreatFeed',
        'fsdb': 'gawseed.threatfeed.feeds.fsdb.FsdbThreatFeed',
    },
    DATASOURCE_KEY: {
        'fsdb': 'gawseed.threatfeed.datasources.fsdb.FsdbDataSource',
        'bro': 'gawseed.threatfeed.datasources.bro.BroDataSource',
        'json': 'gawseed.threatfeed.datasources.json.JsonDataSource',
        'kafka': 'gawseed.threatfeed.datasources.kafka.KafkaDataSource',
        'druid': 'gawseed.threatfeed.datasources.druid.DruidDataSource',
    },
    SEARCHER_KEY: {
        'ssh': 'gawseed.threatfeed.search.ssh.SSHSearch',
        'ip': 'gawseed.threatfeed.search.ip.IPSearch',
        'iprange': 'gawseed.threatfeed.search.iprange.IPRangeSearch',
        'http': 'gawseed.threatfeed.search.http.HTTPSearch',
        'dns': 'gawseed.threatfeed.search.dns.DNSSearch',
        're': 'gawseed.threatfeed.search.re.RESearch',
        'parallel': 'gawseed.threatfeed.search.parallel.Parallel',
    },
    REPORTER_KEY: {
        'dumper': 'gawseed.threatfeed.events.dumper.EventStreamDumper',
        'printer': 'gawseed.threatfeed.events.printer.EventStreamPrinter',
        'reporter': 'gawseed.threatfeed.events.reporter.EventStreamReporter',
        'archiver': 'gawseed.threatfeed.events.archiver.ArchiveReporter',
    },
    ENRICHMENT_KEY: {
        'url': 'gawseed.threatfeed.enrichments.EnrichmentURL',
        'datasource': 'gawseed.threatfeed.enrichments.datasource.Datasource',
        'sorter': 'gawseed.threatfeed.enrichments.sorter.EnrichmentSort',
        'summarizer': 'gawseed.threatfeed.enrichments.summarizer.Summarizer',
        'dnssummarizer': 'gawseed.threatfeed.enrichments.dnssummarizer.DNSSummarizer',
        'connectioncounter': 'gawseed.threatfeed.enrichments.connectionCounter.ConnectionCounter',
        'connectiongrapher': 'gawseed.threatfeed.enrichments.connectionGrapher.ConnectionGrapher',
        'similaripsgraph': 'gawseed.threatfeed.enrichments.similarIPsGraph.SimilarIPsGraph',
        'prioritytotal': 'gawseed.threatfeed.enrichments.prioritySum.PriorityTotal',
    },
}

class Loader():
    YAML_KEY = YAML_KEY
    THREATSOURCE_KEY = THREATSOURCE_KEY
    DATASOURCE_KEY = DATASOURCE_KEY
    SEARCHER_KEY = SEARCHER_KEY
    REPORTER_KEY = REPORTER_KEY
    ENRICHMENT_KEY = ENRICHMENT_KEY
    YAML_SECTIONS = YAML_SECTIONS
    MODULE_XFORMS = MODULE_XFORMS

    def __init__(self):
        self._named_entries = {}

    def dump_config_options(self, debug=False):
        print("threat-search:")
        first_char = "-"
        for part in MODULE_XFORMS:
            print("  %s %s:" % (first_char, part))
            print("    # ---- %s modules and options" % (part))
            print("    # (pick one module)")
            for module in MODULE_XFORMS[part]:
                print("      module: %s" % (module))

                try:
                    module = self.load_module_name(MODULE_XFORMS[part][module])
                    # this forces a module to just dump out config settings to stdout
                    if module.__doc__:
                        doc = module.__doc__
                        doc = re.sub("\n", "\n      # ", doc)
                        print("      #       " + doc)
                        
                    if part == SEARCHER_KEY:
                        module({'dump_config': 1}, None, None, False)
                    elif part == ENRICHMENT_KEY:
                        module({'dump_config': 1}, None, None, False)
                    else:
                        module({'dump_config': 1})
                except Exception as e:
                    print("      # couldn't get config for this")
                    if debug:
                        print("      # " + str(e))
                        print(traceback.format_exc())
                        

                print("")
                first_char = " "
        exit()
        
    def load_module_name(self, module_name):
        """Given a module name, split it in parts, load the module and find its child name
        within and return the final loaded object"""
        try:
            lastdotnum = module_name.rindex('.')
            modulename = module_name[:lastdotnum]   # just the module name
            class_name = module_name[lastdotnum+1:] # just the class init function name

            module = importlib.import_module(modulename)
        except Exception as e:
            print(e)
            raise ValueError("Error parsing/loading module/class name: " + module_name)

        if not hasattr(module, class_name):
            raise ValueError("Error finding class name '" + str(class_name) + "' in module '" + str(module) + "'")

        return getattr(module, class_name)

    def load_yaml_config(self, config_stream, config_parameters):
        """Loads a yaml config from a specific stream and loads class definitions based on it."""

        config_vars = {}
        for item in config_parameters:
            try:
                (name, value) = item.split("=")
                config_vars[name] = value
            except:
                raise ValueError("illegal parameter/syntax passed: {param}".format(item))

        conf_data = config_stream.read()
        template = jinja2.Environment(loader=jinja2.FileSystemLoader("./")).from_string(conf_data)

        conf_data = template.render(config_vars)
        conf = yaml.load(conf_data, Loader=yaml.FullLoader)
        
        self.load_class_configs(conf[YAML_KEY])
        return conf

    def load_class_config(self, section, part):
        module = part['module']
        if module in MODULE_XFORMS[section]:
            module = MODULE_XFORMS[section][module]
        part['class'] = self.load_module_name(module)

    def load_class_configs(self, threatconfs, sections=YAML_SECTIONS):
        """Loops through config and loads classes for each one into a 'class' key"""

        # load each section
        for threatconf in threatconfs:
            for section in sections:
                if section in threatconf: # some are optional
                    parts = threatconf[section]
                    if type(parts) != list:
                        parts = [parts]
                    for part in parts:
                        self.load_class_config(section, part)

        return threatconf

    def maybe_save_entry(self, conf):
        """Remembers a named configuration template for use later"""
        if 'name' not in conf:
            return

        entry_name = conf['name']
        self._named_entries[entry_name] = dict(conf)
        del self._named_entries[entry_name]['name']
        
    def copy_entry(self, entry_name, conf):
        """Creates a new configuration dictionary based on a named template and override values."""
        if entry_name not in self._named_entries:
            return conf

        # clone the template source
        newconf = dict(self._named_entries[entry_name])

        # add/override other variables
        for item in conf:
            newconf[item] = conf[item]

        return newconf

    def create_instance(self, subconf, module_type, args=[], initialize=True):
        """Creates an instantiated instance of a threat-feed module"""

        # save this if requested
        self.maybe_save_entry(subconf)
            
        # if they've requested a template
        if 'use' in subconf:
            subconf = self.copy_entry(subconf['use'], subconf)

        if 'class' not in subconf:
            self.load_class_config(module_type, subconf)
        
        obj = subconf['class']

        created = obj(subconf, *args)

        if initialize:
            created.initialize()

        return created


    def create_instance_for_module(self, conf, module_type, args=[], initialize=True):
        subconf = conf[module_type]
        return self.create_instance(subconf, module_type, args, initialize)



from gawseed.threatfeed.search import Search

class IPSearch(Search):
    """Searches for IP address threats in multiple search_keys fields"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)
        self._search_keys = self.config('search_keys', ['id_orig_h', 'id_resp_h'],
                                        help="A list of fields to search for IP addresses in the data stream")

    def initialize(self):
        super().initialize()
        self._search_keys = self._data_iterator.encode_list(self._search_keys)
        self._data_iterator.set_hints(self._search_keys, self._search_list)
    
    def search(self, row):
        for key in self._search_keys:
            if key not in row:
                continue
            if row[key] in self._search_list:
                return self._search_list[row[key]]
        return None




"""A druid datasource module for the gawseed threat-feed project"""
import time
from pydruid.db import connect

from . import DataSource

class DruidDataSource(DataSource):
    """Connects to a druid server to query for records."""

    def __init__(self, conf):
        super().__init__(conf)

        self.require(['host', 'table'])

        self._host = self.config('host')
        self._port = self.config('port', '8888')
        self._path = self.config('path', '/druid/v2/sql')
        self._scheme = self.config('scheme', 'http')
        self._begin_time = self.config('begin_time', datatype='time')
        self._end_time = self.config('end_time', datatype='time')
        self._limit = self.config('limit', 10000)
        self._table = self.config('table')
        self._where = self.config("where")
        self._sql = self.config('sql')

        self._conn = None
        self._curs = None

    def format_unix_epoch(self, etime):
        """Converts a unix epoch timestamp into a druid comparable date
        string."""
        (tm_year, tm_mon, tm_mday, tm_hour, tm_min,
         tm_sec, tm_wday, tm_yday, tm_isdst) = time.gmtime(etime)
        formatted_time = "%04d-%02d-%02d %02d:%02d:%02d"
        formatted_time = formatted_time % (tm_year, tm_mon, tm_mday,
                                           tm_hour, tm_min, tm_sec)
        return formatted_time

    def set_hints(self, keynames, hints):
        if not self._where:
            clauses = []
            for keyname in keynames:
                for key in hints:
                    clauses.append("%s = '%s'" % (keyname, key)) # XXX: quote escaping/filtering
            self._where = " and ".join(clauses)

    def open(self):
        """Establishes a connection to the druid server."""
        self._conn = connect(host=self._host, port=self._port,
                             path=self._path, scheme=self._scheme)
        self._curs = self._conn.cursor()

        if not self._sql:
            # build the sql from our parameters
            self._sql = "select * from %s where " % (self._table)

            if self._begin_time:
                self._sql += "__time >= time_parse('%s') and " % \
                    (self.format_unix_epoch(self._begin_time))

            if self._end_time:
                self._sql += "__time <= time_parse('%s') and " % \
                    (self.format_unix_epoch(self._end_time))

            for item in self._exclude_list:
                self._sql += "%s <> '%s' and " % \
                    (self._exclude_key, item)
                

            if self._where:
                self._sql += " (" + self._where + ") and "

            if self._sql[-4:] == "and ": # strip off the trailing and
                self._sql = self._sql[:-4]

            if self._limit:
                self._sql = self._sql + "limit %d " % (self._limit)

        self._curs.execute(self._sql)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._curs)._asdict()



import re
import ipaddress
from bisect import bisect

from gawseed.threatfeed.search.ip import IPSearch

class IPRangeSearch(IPSearch):
    """Searches for IP address range threats in multiple search_keys fields"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)
        self._search_keys = self.config('search_keys', ['id_orig_h', 'id_resp_h'],
                                        help="A list of fields to search for IP addresses in the data stream")

    def initialize(self):
        super().initialize()
        self.initialize_ranges()

    def ip_to_int(self, ip):
        if ip.find(":") != -1:
            # ipv6
            return (int(ipaddress.IPv6Address(ip)), 128)
        else:
            return (int(ipaddress.IPv4Address(ip)), 32)

    def initialize_ranges(self):
        self._range_list = []
        self._left_keys = []

        # modify the lists to be ranges
        reblock = re.compile("([^/]+)/([0-9]+)$") # no netmasks currently
        for search_item in self._search_list:
            # is it a 2-column range?
            if type(search_item) == list:
                (ip_lft, mask) = self.ip_to_int(search_item[0])
                (ip_rht, mask) = self.ip_to_int(search_item[1])
                self._left_keys.append(ip_lft)
                self._range_list.append({"left": ip_lft,
                                         'right': ip_rht,
                                         'match': search_item})

            # or it should be a address/netmask format
            else:
                results = reblock.search(search_item)
                if results:
                    # get the mask and mask size_value for v4/v6
                    (ip, masksize) = self.ip_to_int(results.group(1))
                    mask = masksize-int(results.group(2))

                    # get the integer start/end of the range
                    ip_lft = ip - ip % (2**mask)
                    ip_rht = ip_lft + 2**mask-1

                    # save them
                    self._left_keys.append(ip_lft)
                    self._range_list.append({"left": ip_lft,
                                             'right': ip_rht,
                                             'match': search_item})
                else:
                    raise ValueError("entry %s is an unparsable range" % (search_item))
                pass

        # need to sort them into roughly the right order (overlaps dealt with later)
        self._left_keys = sorted(self._left_keys)
        self._range_list = sorted(self._range_list, key=lambda x: x['left'])
        self._length = len(self._left_keys)

    def search(self, row):
        # for each key we want to search for
        for key in self._search_keys:
            # extract the ip address for the key
            (ip, mask) = self.ip_to_int(row[key])
            # see if that address is somewhere within our list of addresses
            point = bisect(self._left_keys, ip)
            if point != 0:
                range_info = self._range_list[point-1]
                if ip < range_info['left'] or ip > range_info['right']:
                    # no match at all; we fell between ranges
                    return None

                # found a match, but see if we have a better match
                # left or right
                range_answer = range_info

                # see if we have a more exact answer
                left_point = point - 1
                if left_point < 0:
                    return range_answer
                
                range_info = self._range_list[left_point]
                while left_point >= 0 and \
                      ip >= range_info['left'] and \
                      ip <= range_info['right']:

                    # see if the new range is more narrow than current
                    if range_info['right']-range_info['left'] < \
                       range_answer['right']-range_answer['left']:
                        range_answer = range_info

                    left_point -= 1
                    range_info = self._range_list[left_point]

                # see if we have a more exact answer
                right_point = point + 1
                if right_point >= self._length:
                    return range_answer
                
                range_info = self._range_list[right_point]
                while right_point < self._length and \
                      ip >= range_info['left'] and \
                      ip <= range_info['right']:

                    # see if the new range is more narrow than current
                    if range_info['right']-range_info['left'] < \
                       range_answer['right']-range_answer['left']:
                        range_answer = range_info

                    right_point += 1
                    range_info = self._range_list[right_point]


                return range_answer

        return None




import sys
import jinja2
import yaml

from gawseed.threatfeed.events import EventStream

class ArchiveReporter(EventStream):
    """Archives the count/row/match/enrichments to a python structure with
    the same key names"""
    def __init__(self, conf):
        super().__init__(conf)

        self._format = self.config('format', 'pickle',
                                   help="The output format to use (currently only pickle).")

    def initialize(self):
        self._dumps = None
        if self._format == 'pickle':
            import pickle
            self._dumps = pickle.dumps
            self._output_type = "wb"
        elif self._format == 'json':
            import json
            self._dumps = json.dumps
        else:
            raise ValueError("Unsupported archive format: " + self._format)

    def write_row(self, count, row, match, enrichments):
        output = self._dumps({ 'count': count,
                               'row': row,
                               'match': match,
                               'enrichments': enrichments})
        self.output(output)




import sys

from gawseed.threatfeed.config import Config

class EventStream(Config):
    def __init__(self, conf):
        super().__init__(conf)
        stream = self.config('stream')
        if type(stream) == str:
            if stream == "stdout":
                self._stream = sys.stdout
            elif stream == "stderr":
                self._stream = sys.stderr
            else:
                self._stream_pattern = stream
        else:
            self._stream_pattern = None
            if stream == None:
                self._stream = sys.stdout
            else:
                self._stream = stream # assume an opened filehandle

        self._output_type = "w"

    def new_output(self, count, **kwargs):
        if self._stream_pattern:
            if self._stream_pattern.find("%d") != -1:
                filename = self._stream_pattern % (count)
            else:
                filename = self._stream_pattern.format(count=count, **kwargs)

            self._stream = open(filename, self._output_type)

    def maybe_close_output(self):
        if self._stream_pattern:
            self._stream.close()

    def output(self, something):
        self._stream.write(something)

    def maybe_convert(self, row):
        new_row = {}
        for key in row:
            value = row[key]
            if type(key) == bytes:
                key = key.decode()
            if type(value) == bytes:
                new_row[key] = value.decode()
            else:
                new_row[key] = value
        return new_row

    def write(self, count, row, match, enrichments):
        row = self.maybe_convert(row)
        self.write_row(count, row, match, enrichments)
        self.verbose("created output for event %d" % (count,))



import tempfile
import graphviz
import re
from gawseed.threatfeed.config import Config

class ConnectionGrapher(Config):
    """Summarizes connection information in BRO or similar data"""
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf)

        self.require(['enrichment_keys', 'output_key'])

        self._enrichment_keys = self.config("enrichment_keys",
                                            datatype=list,
                                            help="The enrichment keys for the data to be graphed")
        self._output_dir = self.config('output_dir', '/tmp',
                                       help="The name of the output directory to store the generated png in")
        self._output_key = self.config('output_key', 'datasource',
                                       help="The output key to store the generated png file name in")
        self._output_type = self.config('output_type', 'png',
                                        help="The output format of the generated file.")

        self._renderer = self.config('renderer', 'fdp',
                                     help='The graphviz renderer to use')
        self._limit = self.config('limit', 100,
                                  help="Don't plot more than LIMIT edges")
        self._minbytes = self.config('minbytes', 0,
                                     help="don't show connections with less than this number of rx or tx bytes")
        self._highlight = self.config('highlight', ['id_orig_h', 'id_resp_h'],
                                      help="Key names from the matched data to highlight")

        self._seen = {}

    def add_node(self, dot, row, label):
        if label in self._seen:
            return self._seen[label]
        
        if label in self._highlight_values:
            self._seen[label] = dot.node(label, color="blue",
                                         fillcolor="lightblue",
                                         style='filled',
                                         shape="box")
        else:
            self._seen[label] = dot.node(label)

    def gather(self, count, row, match, enrichment_data):
        """Re-sort all the enrichment data based on the specified column"""
        # extract the current data

        num = 0
        dot = graphviz.Digraph(engine=self._renderer)

        # pull out the highlight values to check
        self._highlight_values = []
        for h in self._highlight:
            self._highlight_values.append(row[h])


        for enrichment_key in self._enrichment_keys:
            if enrichment_key not in enrichment_data:
                self.verbose("key '" + enrichment_key +
                             "' is not in the enrichment data")
                continue

            data = enrichment_data[enrichment_key]['connections']
            # build the graph via graphviz
            try:
                for orig in data:
                    for dest in data[orig]:
                        dest_created = False
                        for port in data[orig][dest]:
                            if data[orig][dest][port]['rxbytes'] > self._minbytes or \
                               data[orig][dest][port]['txbytes'] > self._minbytes:


                                self.add_node(dot, row, orig)
                                self.add_node(dot, row, dest)

                                label = ("%s:%s\nrx=%s\ntx=%s" % \
                                         (port,
                                          str(data[orig][dest][port]['count']),
                                          str(data[orig][dest][port]['rxbytes']),
                                          str(data[orig][dest][port]['txbytes'])))
                                dot.edge(orig, dest,
                                         label=label)
                                num += 1
                                if num > self._limit:
                                    raise ValueError("too many edges")

            except Exception as e:
                print(e)
                # XXX: we'll continue; think about right thing here
                        
                    
        # create a temporary file to store results in
        (fh, name) = tempfile.mkstemp(dir=self._output_dir,
                                      suffix="." + self._output_type)
        fh.close()

        # graphviz force-adds a suffix, so we remove ours before passing 
        prefixname = re.sub("." + self._output_type + "$", "", name)
        dot.render(prefixname, self._output_dir,
                   format=self._output_type)


        return (self._output_key, name)



import pyfsdb
from gawseed.threatfeed.feeds import ThreatFeed

class FsdbThreatFeed(ThreatFeed):
    """Loads a threat data list from a 'key' column in a FSDB formatted file (see pyfsdb)1"""
    def __init__(self, config):
        super().__init__(config)
        self.require(['file', 'key'])
        self._fsdb_file = self.config('file',
                                      help="The file name to read the bro data stream from")

    def open(self):
        self._tfh = pyfsdb.Fsdb(self._fsdb_file,
                                return_type=pyfsdb.RETURN_AS_DICTIONARY)
        self._tfh_index_column = self._tfh.get_column_number(self._value_column)

        if self._begin_time or self._end_time:
            self._tfh_time_column = self._tfh.get_column_number(self._time_column)

        if self._begin_time:
            # find the start of the data based on time
            while True:
                row = next(self)
                if row[self._tfh_time_column] >= self._begin_time:
                    break

    def __iter__(self):
        return self

    def next_row(self):
        return next(self._tfh)

    def column_names(self):
        return self._tfh.column_names

    def index_column_number(self):
        return self._tfh.get_column_number(self._value_column)

    def read(self, max_records = None):
        array = []
        dictionary = {}
        for (count,entry) in enumerate(self._tfh):
            
            if self.maybe_drop_entry(entry, self._value_column):
                continue

            array.append(entry)
            dictionary[entry[self._value_column]] = entry # note, may erase older ones; build array?
            if max_records and count+1 >= max_records:
                break

        return (array, dictionary)

        



from kafka import KafkaConsumer
from kafka.structs import TopicPartition
from msgpack import unpackb

from . import DataSource

class KafkaDataSource(DataSource):
    """Loads data from a Kafka data stream"""
    def __init__(self, conf):
        super().__init__(conf)

        self.require(['bootstrap_servers', 'topic'])

        self._bootstrap_servers = self.config('bootstrap_servers',
                                              help="A list of kafka bootstrap servers to query")
        self._begin_time = self.config('begin_time', datatype='time',
                                       help="The time to start searching from; no value will mean end of stream")
        self._end_time = self.config('end_time', datatype='time',
                                     help="The time to stop a search when reading; no value will mean don't stop streaming")
        self._over_time = self.config('over_time', default="0m", datatype='offset',
                                      help="The amount of time to search before and after desired date in case records aren't organized well by kafka timestamps")
        self._time_column = self.config('time_column',
                                        help="Time column to use when searching through data")
        self._topic = self.config('topic',
                                  help="The kafka topic to search")
        self._consumer_timeout_ms = self.config('timeout',
                                                help="A timeout in milliseconds to wait for server data.")
        self._max_records = self.config('max_records',
                                        help="The maximum number of records to return")

    def initialize(self):
        super().initialize()
        if self._time_column:
            self._time_column = self.encode_item(self._time_column)
        if self._end_time:
            self._kafka_end_time = self._end_time + self._over_time

    def open(self):
        consumer = KafkaConsumer(bootstrap_servers=self._bootstrap_servers)
                                 #consumer_timeout_ms=self._consumer_timeout_ms)
        # point to what we want at
        partition = TopicPartition(self._topic,0)
        consumer.assign([partition])

        if self._begin_time:
            offinfo = consumer.offsets_for_times({partition: self._begin_time * 1000 - self._over_time})
            if offinfo == None or offinfo[partition] == None:
                raise ValueError("There is no data in the enterprise stream the begin date")
            offset = offinfo[partition].offset
            consumer.seek(partition, offset)

        self._consumer = consumer
        self.verbose("opened kafka datasource: " + str(self.get_config()))
        return self

    def __iter__(self):
        return self

    def __next__(self):
        row = next(self._consumer)
        decoded_row = unpackb(row.value)
        if self._end_time:
            # self.verbose("searching forward from:")
            # self.verbose(decoded_row)
            count = 0
            while True:
                count += 1
                decoded_time = decoded_row[self._time_column]
                decoded_time = self.decode_item(decoded_time)
                decoded_time = self.parse_time(decoded_time)
                if decoded_time >= self._kafka_end_time:
                    self.verbose("kafka end time reached: " + str(count)+ " rows")
                    raise StopIteration()

                # see if it's within the time window
                # and that it's not in the filter list
                if decoded_time >= self._begin_time \
                   and decoded_time <= self._end_time \
                   and entry[self._exclude_column] in self._exclude_list:
                    # self.verbose("row found after" + str(count)+ " rows")
                    return decoded_row

                # else continue searching for a row that does match
                row = next(self._consumer)
                decoded_row = unpackb(row.value)
                
        return decoded_row

    def default_is_binary(self):
        return True





from pyfsdb.Fsdb import RETURN_AS_DICTIONARY,Fsdb

from gawseed.threatfeed.datasources.fsdb import FsdbDataSource

class BroDataSource(FsdbDataSource):
    """Loads data from BRO text-based (tab separated) log files"""
    def __init__(self, conf):
        super().__init__(conf)
        self._file_handle = self.config('file_handle', datatype='file_handle',
                                        help="A python3 opened file handle for the BRO data to be streamed")
        self._file = self.config('file',
                                 help="The file name to read the bro data stream from")

        self._begin_time = self.config('begin_time', datatype='time',
                                       help="The time to start searching from; no value will mean end of stream")
        self._end_time = self.config('end_time', datatype='time',
                                     help="The time to stop a search when reading; no value will mean don't stop streaming")
        self._time_column = self.config('time_column',
                                        help="Time column to use when searching through data")

    def open(self):
        if self._file and not self._file_handle:
            self._file_handle = open(self._file, "r")
        
        column_names = None
        for line in self._file_handle:
            if line[0] != "#": # skip at end of headers
                break

            if line[0:7] == "#fields":
                column_names = line.replace(".", "_").split("\t")
                column_names.pop(0)

        if not column_names:
            raise ValueError("passed bro file isn't in expected bro file format")

        import pdb ; pdb.set_trace()
        self._fh = Fsdb(file_handle=self._file_handle,
                        return_type=RETURN_AS_DICTIONARY)
        self._fh.column_names = column_names
        # todo:: XXX: we eat a line here; fix this

        self.maybe_skip_to_time()

    def initialize(self):
        super().initialize()



import sys

from gawseed.threatfeed.events import EventStream

class EventStreamPrinter(EventStream):
    """Prints simple summaries of events found."""
    def __init__(self, conf):
        super().__init__(conf)
        self._form = self.config('format', "  %-30.30s: %s\n",
                                 help="The line format to use when printing")
        self._extra_fields = self.config('extra_fields', [],
                                         help="Extra fields to include in the output")
        self._timestamp = self.config('timestamp','ts',
                                      help="The column name of the timestamp data")

    def out(self, info1, info2):
        self.output(self._form % (str(info1) + ":", str(info2)))

    def out_if(self, info1, row, key):
        if key in row:
            self.out(info1, row[key])

    def write_row(self, count, row, match, enrichments):
        self.output("match #" + str(count) + ":\n")

        self.output("Threat information:\n")
        self.out_if("Match value", match, 'value')
        self.out_if("Tag", match, 'tag')

        self.output("Data source information:\n")
        self.out_if("Timestamp", row, self._timestamp)
        self.out_if("id_orig_h", row, 'id_orig_h')
        self.out_if("id_resp_h", row, 'id_resp_h')

        if len(self._extra_fields) > 0:
            self.output("Extra information:\n")
            for field in self._extra_fields:
                self.out_if(field, row, field)

        self.output("-" * 70 + "\n")




from gawseed.threatfeed.enrichments import Enrichment

class HAASIPEnrichment(Enrichment):
    def __init__(self, conf):
        super().__init__(conf)
        self.require(['url'])

        self._url = self.config('url')
        self._tag = self.config('tag')

    def type():
        return 'application/json'

    def gather(self, key):
        return self.geturl(self._url.format(tag=self._tag, key=key)



from gawseed.threatfeed.config import Config

class DataSource(Config):
    def __init__(self, conf):
        super().__init__(conf)

        self._binary = self.config("binary", self.default_is_binary(),
                                   help="If the datasource is binary (bytes) data, set this to convert when necessary")
        self._exclude_column = self.config('exclude_key', 'value',
                                           help="The primary column/key name to use for checking whether to exclude data")

        self._exclude_list = self.config('exclude', [],
                                         help='A list of entries to ignore in the threat feed')
        

    def initialize(self):
        super().initialize()

    def default_is_binary(self):
        return False

    def is_binary(self):
        return self._binary

    def convert_row_to_utf8(self, row):
        if not self._binary:
            return row
            
        utf8_row = {}
        for item in row:
            utf8_row[self.decode_item(item)] = self.decode_item(row[item])
        return utf8_row

    def encode_item(self, item):
        """converts a string item to a utf-8 bytes version of it"""

        # don't encode if not needed
        if not self._binary:
            return item

        if type(item) == bytes:
            return item
        return bytes(item, 'utf-8')

    def decode_item(self, item):
        """Decodes a bytes object into a utf-8 string"""

        # don't decode if not needed
        if not self._binary:
            return item

        if type(item) == bytes:
            return item.decode()

        if type(item) == list:
            return self.decode_list(item) 

        return item

    def encode_dict(self, old_dict):
        """Creates a new dict that contains both binary and string based indexes for all entries"""

        # don't encode if not needed
        if not self._binary:
            return old_dict

        new_dict = {}
        for key in old_dict:
            # stores both new binary key and the old
            new_dict[self.encode_item(key)] = old_dict[key]
            new_dict[key] = old_dict[key]
        return new_dict

    def encode_list(self, old_list):
        """Creates a new  list converted to binary if necessary"""

        # don't encode if not needed
        if not self._binary:
            return old_list

        new_list = []
        for key in old_list:
            # stores both new binary key and the old
            new_list.append(self.encode_item(key))
        return new_list

    def decode_list(self, old_list):
        new_list = []
        for key in old_list:
            # stores both new binary key and the old
            new_list.append(self.decode_item(key))
        return new_list

    def maybe_convert_token_to_binary(self, value):
        # don't encode if not needed
        if not self._binary:
            return value

        if type(value) != str:
            return value
        if self._binary_search:
            return bytes(value, 'utf-8')
        return value

    def maybe_convert_list_to_binary(self, values):
        # don't encode if not needed
        if not self._binary:
            return values

        new_list = []
        for value in values:
            new_list.append(self.maybe_convert_token_to_binary(value))
        return new_list
    
    def set_hints(self, keynames, hint_dict):
        """Sets a set of hints for things to look for; datasources that can
        more easily sub-select can use these hints to refine their
        data retrieval.  Datasources that can not will simply ignore
        the hints, and thus searchers should double check the
        results.
        """
        pass

    def close(self):
        pass


    def maybe_drop_entry(self, entry):
        """Returns false if the entry should not be dropped (ie, should be
        searched)"""
        if self._exclude_column not in entry:
            return False

        if entry[self._exclude_column] in self._exclude_list:
            return True

        return False
        
    



import json
import time
import sys
import urllib3

from gawseed.threatfeed.config import Config

class EnrichmentURL(Config):
    """Pulls enrichment data from a supplied URL.  The URL format line will be passed a 'tag' and 'match_info'."""
    def __init__(self, conf, search_index, data_source, is_binary, loader=None):
        super().__init__(conf)
        
        self._pool = urllib3.PoolManager()
        self._conf = conf
        self._cache = {}
        self._loader = loader

        self.require(['url'])

        self._url = self.config('url',
                                help="A URL to use for gathering enrichment data.  It may be a python string to be formatted with a 'tag' and 'match_info' value")
        self._tag = self.config('tag',
                                help="The tag to be passed to the URL")
        self._type = self.config('type','text/plain',
                                 help="The expected content type to be returned.  Non-matches will be dropped")
        self._match_key = self.config('match_key', 'value',
                                      help="The match key column name to be used")
        self._data_key = self.config('data_key', 'key',
                                     help="The data key to used")
        self._output_key = self.config('output_key', 'geturl',
                                       help="The output key to store the returned data in.")

        self._cache_time = self.config('cache_time', 3600,
                                       help="If set, will cache the data for a particular key for this number of seconds")

    def geturl(self, url, reqtype='GET', params={}):
        prior_results = self.check_cache(url)
        if prior_results:
            return prior_results
        
        self.verbose("Fetching URL: " + url)
        r = self._pool.request(reqtype, url)
        if r.status != 200:
            print("  failed to fetch URL:" + str(r.status))
            return None

        # XXX: check against the expected type (self._type)
        return self.cache(url, r.data)

    def convert(self, result):
        if self._type == 'text/plain' or self._type == 'text':
            return result.decode()
        if self._type == 'application/json' or self._type == 'json':
            return json.loads(result.decode())
        
        # (probably shouldn't do this)
        return result
        
    def gather(self, count, row, match, enrichment_data = {}):
        fetched = self.geturl(self._url.format(tag=match['tag'],
                                               match_info=match[self._match_key],
                                               data_info=row[self._data_key]))
        if not fetched:
            sys.stderr.write("Failed to gather enrichment data...\n")
            return (None, None)
        converted = self.convert(fetched)
        return (self._output_key, converted)

    def cache(self, key, value):
        if not self._cache_time:
            return value
        self._cache[key] = { 'data': value,
                             'timestamp': time.time() }
        return value
        
    def check_cache(self, key):
        if not self._cache_time:
            return None

        if key in self._cache and time.time() < self._cache[key]['timestamp'] + self._cache_time:
            return self._cache[key]['data']

        return None



import sys

from gawseed.threatfeed.events import EventStream

class EventStreamDumper(EventStream):
    """A debugging event reporter: just outputs python strings of matches found"""
    def __init__(self, conf):
        super().__init__(conf)

        self._pretty_print = self.config("pretty", default=False,
                                         help="Pretty print the results")

    def initialize(self):
        if self._pretty_print:
            import pprint
            self._printer = pprint.pformat
        else:
            self._printer = str

    def write(self, count, row, match, enrichments):
        self.output("match #" + self._printer(count) + ":\n")
        self.output("row: " + self._printer(row) + "\n")
        self.output("--\n")
        self.output("match: " + self._printer(match) + "\n")
        self.output("--\n")
        self.output("enrichments: " + self._printer(enrichments) + "\n")
        self.output("--------------")



from gawseed.threatfeed.config import Config

class EnrichmentSort(Config):
    """Sorts an existing enrichment dataset and puts it back"""
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf)

        self.require(['enrichment_key'])

        self._enrichment_key = self.config("enrichment_key",
                                           help="The enrichment key for the data to be sorted")
        self._sort_key = self.config("sort_key", "ts",
                                     help="Sort the enrichment data by this key")

    def gather(self, count, row, match, enrichment_data):
        """Re-sort all the enrichment data based on the specified column"""
        # extract the current data
        if self._enrichment_key not in enrichment_data:
            return (None, None)

        data = enrichment_data[self._enrichment_key]
        data = sorted(data, key=lambda x: x[self._sort_key])
        enrichment_data[self._enrichment_key] = data

        return (self._output_key, data)



import sys
import re

from gawseed.threatfeed.search.re import RESearch

class HTTPSearch(RESearch):
    """Matches a regexp threat against a host/uri combo data field (aka bro log)"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)

        self._host_key = self.config('host_key', 'host',
                                     help="The column name for the host to construct a URL out of")
    
        self._uri_key = self.config('uri_key', 'uri',
                                     help="The column name for the uri portion of the URL")

    def initialize(self):
        super().initialize()
        self._auth_success_key = self._data_iterator.encode_item(self._host_key)
        self._auth_success_value = self._data_iterator.encode_item(self._uri_key)

    def search(self, row):
        # XXX: currently assumes a split host/uri scheme like bro.  Need to make this generic
        if self._host_key not in row or self._uri_key not in row or row[self._host_key] is None or row[self._uri_key] is None:
            return None

        host = self._data_iterator.decode_item(row[self._host_key])
        uri = self._data_iterator.decode_item(row[self._uri_key])

        url = "http://" + host + "/" + uri

        return self.search_one(url)




from gawseed.threatfeed.config import Config

class PriorityTotal(Config):
    """Combines priorities found in the base match, and all enrichment
    data.  Uses the 'priority_adj' field as a search criteria.
    """
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf, search_index, dataset, is_binary, loader)
        
        self._output_key = self.config('output_key', 'priority_total',
                                       help="The output key to store the summarized priority in.")

        self._search_key = self.config('search_key', 'priority_adj',
                                       help="The search key to use when finding priority adjustment fields.")

    def gather(self, count, row, match, enrichment_data):
        """Sum the various priority fields together for a final value."""

        priority = 0 # starting value

        if 'priority' in match:
            priority = match['priority'] # unlikely
        elif 'priority' in row:
            priority = row['priority'] # also unlikely
            
        if not self.check_enrichment_data(enrichment_data):
            return (None, None)

        for key in enrichment_data:
            edata = enrichment_data[key]
            if isinstance(edata, dict):
                edata = [edata]
            elif not isinstance(edata, list):
                continue # unknown type to search
            
            for data in edata:
                if isinstance(data, dict):
                    if self._search_key in data:
                        priority += edata[self._search_key]
                
        return (self._output_key, {
            self._output_key: priority
        })
                



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gawseed-threat-feed-tools",
    version="1.1",
    author="Wes Hardaker and USC/ISI",
    author_email="opensource@hardakers.net",
    description="Tools to search network data logs for threat feed data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gawseed/threat-feed-tools",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'threat-search.py = gawseed.threatfeed.main:main',
            'pkl-to-report.py = gawseed.threatfeed.pkl2report:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=['pyfsdb>=0.9.92',
                      'kafka-python',
                      'python-dateutil',
                      'jinja2',
                      'pyyaml>5',
                      'msgpack',
                      'lz4',
                      'graphviz',
                      'dnssplitter',
                      'urllib3'],
    python_requires = '>=3.0',
    test_suite='nose.collector',
    tests_require=['nose'],
)



import pyfsdb
from . import DataSource

class FsdbDataSource(DataSource):
    """Loads data from a tab-separated FSDB formatted file (see pyfsdb)"""
    def __init__(self, conf):
        super().__init__(conf)
        
        self._file_handle = self.config('file_handle', datatype='file_handle',
                                        help="A python3 opened file handle for the BRO data to be streamed")
        self._file = self.config('file',
                                 help="The file name to read the bro data stream from")
        self._begin_time = self.config('begin_time', datatype='time',
                                       help="The time to start searching from; no value will mean end of stream")
        self._end_time = self.config('end_time', datatype='time',
                                     help="The time to stop a search when reading; no value will mean don't stop streaming")
        self._time_column = self.config('time_column',
                                        help="Time column to use when searching through data")

    def initialize(self):
        super().initialize()
        if not self._file_handle and not self._file:
            self.config_error("either file_handle or file is required for the %s module" % (type(self)))

    def open(self):
        self._fh = pyfsdb.Fsdb(file_handle=self._file_handle, filename=self._file,
                               return_type=pyfsdb.RETURN_AS_DICTIONARY)


        self.maybe_skip_to_time()

    def maybe_skip_to_time(self):
        if self._begin_time or self._end_time:
            self._fh_time_column = self._fh.get_column_number(self._time_column)

        if self._begin_time:
            # find the start of the data based on time
            while True:
                row = next(self)
                if float(row[self._time_column]) >= self._begin_time:
                    break

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            row = next(self._fh)
            if self._end_time and float(row[self._time_column]) >= self._end_time:
                raise StopIteration()
            if not self.maybe_drop_entry(row):
                break # don't filter

        return row

    def close(self):
        self._fh.fileh.close()
        self._fh.close()
        



import re
import time
import sys
from dateutil import parser

class Config():
    def __init__(self, config={}):
        self._config = config
        self._number_re=re.compile("^@?[0-9.]+$")
        self._verbose = False
        if 'verbose' in config:
            self._verbose = True

    def initialize(self):
        """Overridable function for doing things beyond config copying in __init__"""
        pass
    
    def set_defaults(self, defaults={}):
        for default in defaults:
            if default not in self_.config:
                self._config = defaults[default]

    def require(self, requirements):
        if type(requirements) != list:
            requirements = [requirements]
        if 'dump_config' in self._config:
            print("      # required: " + str(requirements))
            return

        if 'use' in self._config:
            return # assume they'll pull it from something named

        for requirement in requirements:
            if requirement not in self._config:
                self.config_error("'%s' is a requirement argument for %s" % (requirement, type(self)))

    def config(self, name, default=None, help=None, datatype=None):
        if 'dump_config' in self._config:
            if help:
                print("      # %s: %s" % (name, help))
            print("      %s: %s" % (name, default))
            return

        value = default
        if name in self._config:
            value = self._config[name]

        if value is None:
            return None

        if datatype and datatype == 'time':
            return self.parse_time(value)

        if datatype and datatype == 'offset':
            return self.parse_offset(value)

        if datatype and (datatype == 'list' or datatype == list):
            if type(value) != list:
                value = [value]

        if datatype == 'file_handle' and value == 'stdin':
            value = sys.stdin
        
        return value

    def get_config(self):
        return self._config

    def config_error(self, msg):
        raise ValueError(msg)

    def parse_offset(self, timestr):
        multipliers = {'m': 60,
                       'h': 3600,
                       'd': 3600*24,
                       'w': 3600*24*7,
                       'y': 3600*24*365.25}

        multiplier = 1
        if timestr[-1] in multipliers:
            multiplier = multipliers[timestr[-1]]
            timestr = timestr[:-1]

        value = float(timestr) * multiplier
        return value
        

    def parse_time(self, timestr):
        if timestr[0] == '+' or timestr[0] == '-':
            # return an offset from now
            now = time.time()
            return now + self.parse_offset(timestr)
        elif self._number_re.match(timestr): # assume epoch seconds
            if timestr[0] == '@':
                timestr = timestr[1:]
            return float(timestr)
        else:
            # hope for the best that this can parse it
            return parser.parse(timestr).timestamp()

    def verbose(self, message):
        if self._verbose:
            sys.stderr.write(str(message) + "\n")



from gawseed.threatfeed.config import Config

from collections import Counter

class Summarizer(Config):
    """Summarizes/counts data from another enricher's data;
    Produces a dictionary with data_key/count pairs"""
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf)

        self.require(['enrichment_key', 'data_key'])

        self._data_key = self.config("data_key",
                                     help="The data key to summarize data by")
        self._enrichment_key = self.config("enrichment_key",
                                           help="The enrichment key for the data to be sorted")
        self._output_key = self.config('output_key', 'datasource',
                                       help="The output key to store the returned data in.")

    def check_enrichment_data(self, enrichment_data):
        if self._enrichment_key not in enrichment_data:
            self.verbose("summarizer data wasn't present")
            self.verbose("  keys present:" + str(enrichment_data.keys()))
            self.verbose(self.get_config())
            return False

        if type(enrichment_data[self._enrichment_key]) != list:
            self.verbose("summarizer data wasn't in a list: " + type(enrichment_data[self._enrichment_key]))
            self.verbose(self.get_config())
            return False

        return True


    def gather(self, count, row, match, enrichment_data):
        """Re-sort all the enrichment data based on the specified column"""

        if not self.check_enrichment_data(enrichment_data):
            return (None, None)

        counter = Counter()
        for row in enrichment_data[self._enrichment_key]:
            if self._data_key in row:
                counter[row[self._data_key]] += 1

        return (self._output_key, dict(counter))



import os
import tempfile
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as dates
from gawseed.threatfeed.config import Config

class SimilarIPsGraph(Config):
    """Summarizes connection information in BRO or similar data"""
    def __init__(self, conf, search_index, dataset, is_binary, loader=None):
        super().__init__(conf)

        self.require(['enrichment_key', 'output_key'])

        self._enrichment_key = self.config("enrichment_key",
                                           help="The similar ips key for the data to be graphed")
        self._output_dir = self.config('output_dir', '/tmp',
                                       help="The name of the output directory to store the generated png in")
        self._output_key = self.config('output_key', 'similarips_image',
                                       help="The output key to store the generated png file name in")
        self._output_type = self.config('output_type', 'png',
                                        help="The output format of the generated file.")
        self._match_key = self.config('match_key', 'value',
                                      help="The match key column name to be used")
        self._data_key = self.config('data_key', 'key',
                                     help="The data key to use within the data row")
        self._time_key = self.config('time_key', 'ts',
                                     help="The time key to use within the data row")

        self._seen = {}

    def gather(self, count, row, match, enrichment_data):
        """Plot a time graph of when similar addresses were seen"""
        # extract the current data

        if self._enrichment_key not in enrichment_data:
            self.verbose("key '" + self._enrichment_key +
                         "' is not in the enrichment data")
            return (None, None)

        if enrichment_data[self._enrichment_key]['status'] == 'error':
            return (None, None)
            
        (fh, name) = tempfile.mkstemp(dir=self._output_dir,
                                      suffix="." + self._output_type)
        os.close(fh)

        extra_addresses_x = []
        extra_addresses_y = []

        match_x = []
        match_y = []

        for n, address in enumerate(enrichment_data[self._enrichment_key]['cluster']):
            for timestamp in  enrichment_data[self._enrichment_key]['cluster'][address]['times']:
                tint = int(timestamp)
                if tint <= 0:
                    continue
                if self._data_key and row[self._data_key] == address:
                    match_x.append(tint)
                    match_y.append(n)
                else:
                    extra_addresses_x.append(tint)
                    extra_addresses_y.append(n)

        # create a figure and NxM subplots
        fig, (ax0) = plt.subplots(nrows=1, ncols=1, sharex=True)

        # set various parameters on the first subplot
        ax0.xaxis.set_major_formatter(dates.DateFormatter("%Y/%m/%d\n%H:%M"))
        ax0.set_title('Activity graph')

        if (len(extra_addresses_x)) > 0:
            extra_addresses_x = dates.epoch2num(extra_addresses_x)
            ax0.scatter(extra_addresses_x, extra_addresses_y,
                        s=5, label="Other Addresses")

        if (self._time_key):
            row_stamp = int(self.parse_time(row[self._time_key]))
            row_stamp = dates.epoch2num(row_stamp)
            ax0.plot([row_stamp, row_stamp],
                     [0, len(enrichment_data[self._enrichment_key]['cluster'])],
                     label="Event Time", linewidth=1, ms=1,
                     color="green")

        if (len(match_x)) > 0:
            match_x = dates.epoch2num(match_x)
            ax0.scatter(match_x, match_y,
                        s=5, color="red",
                        label="Matched address")

        plt.ylabel("Enumerated Addresses")
        ax0.legend(loc='best')

        # deals with long labels:
        plt.tight_layout()

        fig.set_dpi(150)
        fig.set_size_inches(11,7.5)
        matplotlib.rcParams.update({'font.size': 10})

        # save it to a file
        plt.savefig(name,
                    bbox_inches="tight", pad_inches=0)

        return (self._output_key, name)



import sys
import jinja2
import yaml

from gawseed.threatfeed.events import EventStream

class EventStreamReporter(EventStream):
    """Formats output based on a jinja2 template.  The template is passed
    the following data fields:
       count: the match number for this run
       row: the row of the data source that was matched
       match: the content from the threat feed that trigger the event
       extra: Any supplemental json data passed in the 'extra_information' config option
       enrichments: A dictionary of any enrichments that were retrieved"""
    def __init__(self, conf):
        super().__init__(conf)

        self.require(['template'])
        self._template = self.config('template',
                                     help="The file name to use as the jinja2 template.")
        self._jinja_extra_information = self.config('extra_information', {},
                                                    help="A YAML file name to be loaded as an extra_information field passed to the jinja2 template")

    def initialize(self):
        self._jinja_template = open(self._template, "r").read()
        self._template = jinja2.Environment(loader=jinja2.FileSystemLoader("./")).from_string(self._jinja_template)


        if self._jinja_extra_information:
            fh = open(self._jinja_extra_information, "r")
            self._jinja_extra_information = yaml.load(fh, Loader=yaml.FullLoader)

    def write_row(self, count, row, match, enrichments):
        output = self._template.render({ 'count': count,
                                         'row': row,
                                         'match': match,
                                         'extra': self._jinja_extra_information,
                                         'enrichments': enrichments})
        self.output(output + "\n")




from gawseed.threatfeed.search.ip import IPSearch

class SSHSearch(IPSearch):
    """Searches data for threats, but requires the auth_success field to
    be True. IE, only successful logins will be considered a match.
    Use the 'ip' module if you don't want this restriction applied."""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        
        super().__init__(conf, search_list, data_iterator, binary_search)
        # Really need a tri-nary option for this

        self._auth_success_key = self.config('auth_success_key', 'auth_success',
                                             help="When searching for authenticated ssh connections, use this column name to determine if authentication suceeded")
        self._auth_success_value = self.config('auth_success_value', True,
                                               help="The value that should match the authentication field identified by auth_success_key")

    def initialize(self):
        super().initialize()
        self._auth_success_key = self._data_iterator.encode_item(self._auth_success_key)
        if type(self._auth_success_value) == str:
            self._auth_success_value = self._data_iterator.encode_item(self._auth_success_value)
    
    def search(self, row):
        if not self._auth_success_key or self._auth_success_key not in row:
            return None
        if row[self._auth_success_key] != self._auth_success_value:
            return None
        return super().search(row)




import sys
import re

from gawseed.threatfeed.search.re import RESearch

class DNSSearch(RESearch):
    """Searches for a threat regexp in DNS query logs"""
    def __init__(self, conf, search_list, data_iterator, binary_search):
        super().__init__(conf, search_list, data_iterator, binary_search)
        self._key = self.config('key', 'query',
                                help="The key of the field in the data stream to search for matches") # assumes bro

    def initialize(self):
        super().initialize()
        self._key = self._data_iterator.encode_item(self._key)

